<?php

declare(strict_types=1);

namespace Drupal\data_migration_report\Commands;

use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Database;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Extension\ExtensionPathResolver;
use Drupal\Core\File\FileExists;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\File\FileUrlGenerator;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\State\StateInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Drupal\data_migration_report\Form\CredentialForm;
use Drupal\field\FieldConfigInterface;
use Drupal\migrate_drupal\MigrationConfigurationTrait;
use Drush\Commands\DrushCommands;
use Symfony\Component\Console\Helper\Table;
use Symfony\Component\Console\Output\ConsoleOutput;
use Symfony\Component\Yaml\Parser;
use Symfony\Component\Yaml\Yaml;

/**
 * A drush command file.
 */
class DataMigrationReportCommands extends DrushCommands {

  use MigrationConfigurationTrait;
  use StringTranslationTrait;

  /**
   * YAML parser.
   *
   * @var \Symfony\Component\Yaml\Parser
   */
  protected Parser $parser;

  /**
   * The source database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected Connection $sourceDatabaseConnection;

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected Connection $database;

  /**
   * Path resolver service.
   *
   * @var \Drupal\Core\Extension\ExtensionPathResolver
   */
  protected ExtensionPathResolver $pathResolver;

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected FileSystemInterface $fileSystem;

  /**
   * The file url generator service.
   *
   * @var \Drupal\Core\File\FileUrlGenerator
   */
  protected FileUrlGenerator $fileUrlGenerator;

  /**
   * The renderer service.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected RendererInterface $renderer;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected EntityFieldManagerInterface $entityFieldManager;

  /**
   * The source data.
   *
   * @var array
   */
  protected array $source = [];

  /**
   * The destination data.
   *
   * @var array
   */
  protected array $dest = [];

  /**
   * The field and column mapping.
   *
   * @var array
   */
  protected array $fieldColumnMapping = [];

  /**
   * The field name mapping.
   *
   * @var array
   */
  protected array $fieldNameMapping = [];

  /**
   * The field preprocess mapping.
   *
   * @var array
   */
  protected array $fieldPreprocessMapping = [];

  /**
   * The field name preprocess mapping.
   *
   * @var array
   */
  protected array $fieldNamePreprocessMapping = [];

  /**
   * Constructs a new DataMigrationReportCommands object.
   *
   * @param \Drupal\Core\Database\Connection $database
   *   The database connection.
   * @param \Drupal\Core\Extension\ExtensionPathResolver $path_resolver
   *   Service used to get the path from module machine name.
   * @param \Drupal\Core\State\StateInterface $state
   *   The state service.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system service.
   * @param \Drupal\Core\File\FileUrlGenerator $fileUrlGenerator
   *   The file url generator service.
   * @param \Drupal\Core\Render\RendererInterface $renderer
   *   The renderer service.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager.
   */
  public function __construct(Connection $database, ExtensionPathResolver $path_resolver, StateInterface $state, FileSystemInterface $file_system, FileUrlGenerator $fileUrlGenerator, RendererInterface $renderer, EntityFieldManagerInterface $entity_field_manager) {
    $this->database           = $database;
    $this->pathResolver       = $path_resolver;
    $this->state              = $state;
    $this->fileSystem         = $file_system;
    $this->fileUrlGenerator   = $fileUrlGenerator;
    $this->renderer           = $renderer;
    $this->entityFieldManager = $entity_field_manager;
  }

  /**
   * Drush Command to run Migration testing.
   *
   * @command generate:content-mapping
   * @usage drush gcm
   * @aliases gcm
   */
  public function generateContentMapping(): void {
    // Check the source database connection and log an error if it's not found.
    $status = $this->checkSourceConnection();

    if (!$status) {
      $this->logger()->error($this->t('No source database connection found.'));
      $this->logger()->error($this->t('Please configure the source database credentials at :config_page', [
        ':config_page' => Url::fromRoute('data_migration_report.migrate_database_credentials', [], ['absolute' => TRUE])->toString(),
      ]));
      return;
    }

    // Generate YAML file for user entity.
    if ($this->getDrupal7ModuleStatus('user')) {
      $fields = $this->contentTypeFields('user', 'user');
      $this->generateMappingYaml('user', 'user', $fields);
    }

    // Generate YAML files for node entities.
    if ($this->getDrupal7ModuleStatus('node')) {
      $node_types = $this->getValues('SELECT type FROM {node_type}', 'd7', 'one_column');

      if (!empty($node_types)) {
        foreach ($node_types as $type) {
          $fields = $this->contentTypeFields('node', $type);
          $this->generateMappingYaml('node', $type, $fields);
        }
      }
    }

    // Generate YAML files for taxonomy_term entities.
    if ($this->getDrupal7ModuleStatus('taxonomy')) {
      $taxonomy_types = $this->getValues('SELECT machine_name FROM {taxonomy_vocabulary}', 'd7', 'one_column');

      if (!empty($taxonomy_types)) {
        foreach ($taxonomy_types as $type) {
          $fields = $this->contentTypeFields('taxonomy_term', $type);
          $this->generateMappingYaml('taxonomy_term', $type, $fields);
        }
      }
    }

    // Generate YAML files for block_custom entities.
    if ($this->getDrupal7ModuleStatus('block')) {
      $fields = $this->contentTypeFields('block_custom', 'block_custom');
      $this->generateMappingYaml('block_custom', 'block_custom', $fields);
    }

    // Generate YAML files for comment entities.
    if ($this->getDrupal7ModuleStatus('comment')) {
      if (!empty($node_types)) {
        foreach ($node_types as $type) {
          $bundle = 'comment_node_' . $type;
          $fields = $this->contentTypeFields('comment', $bundle);
          $this->generateMappingYaml('comment', $bundle, $fields);
        }
      }
    }

    // Generate YAML files for menu_links entities.
    if ($this->getDrupal7ModuleStatus('menu')) {
      $menus = $this->getValues('SELECT menu_name FROM {menu_custom}', 'd7', 'one_column');

      if (!empty($menus)) {
        foreach ($menus as $menu) {
          $this->generateMappingYaml('menu_links', $menu);
        }
      }
    }

    // Generate YAML files for url_alias entities.
    if ($this->getDrupal7ModuleStatus('path')) {
      $this->generateMappingYaml('url_alias', 'url_alias');
    }

    // Generate YAML files for file entities.
    if ($this->getDrupal7ModuleStatus('file')) {
      $this->generateMappingYaml('file', 'file');
      $this->generateMappingYaml('file_private', 'file_private');
    }

    // Generate YAML files for field_collection_item entities.
    if ($this->getDrupal7ModuleStatus('field_collection')) {
      $field_collection_types = $this->getValues("SELECT field_name FROM {field_config} WHERE type='field_collection' AND active=1 AND deleted=0", 'd7', 'one_column');

      if (!empty($field_collection_types)) {
        foreach ($field_collection_types as $type) {
          $fields = $this->contentTypeFields('field_collection_item', $type);
          $this->generateMappingYaml('field_collection_item', $type, $fields);
        }
      }
    }

    // Generate YAML files for paragraphs_item entities.
    if ($this->getDrupal7ModuleStatus('paragraphs')) {
      $paragraphs_bundle = $this->getValues('SELECT bundle FROM {paragraphs_bundle}', 'd7', 'one_column');

      if (!empty($paragraphs_bundle)) {
        foreach ($paragraphs_bundle as $bundle) {
          $fields = $this->contentTypeFields('paragraphs_item', $bundle);
          $this->generateMappingYaml('paragraphs_item', $bundle, $fields);
        }
      }
    }
  }

  /**
   * Returns the status of a Drupal 7 module.
   *
   * This method is used to check if a Drupal 7 module is enabled.
   *
   * @param string $module_name
   *   The name of the module.
   *
   * @return bool
   *   TRUE if the Drupal 7 module is enabled, FALSE otherwise.
   */
  public function getDrupal7ModuleStatus(string $module_name): bool {
    $status = $this->sourceDatabaseConnection->select('system', 's')
      ->fields('s', ['status'])
      ->condition('type', 'module')
      ->condition('name', $module_name)
      ->execute()->fetchField();
    return (bool) $status;
  }

  /**
   * Return destination mapping entity type, bundle and table.
   *
   * @param string $entity_type
   *   The entity type.
   * @param string $bundle
   *   The entity bundle.
   *
   * @return array
   *   The array of destination entity type and bundle.
   */
  public function getDestinationMapping(string $entity_type, string $bundle): array {
    $dest_entity_type = $entity_type;
    $dest_bundle      = $bundle;

    if ($entity_type == 'comment') {
      // The Forum module provides its own comment type (comment_forum).
      if ($bundle == 'comment_node_forum') {
        $dest_bundle = 'comment_forum';
      }
      else {
        // Default generated comment type names, limited to 30 characters.
        $dest_bundle = mb_substr($bundle, 0, 30);
      }
    }

    if ($entity_type == 'field_collection_item') {
      $dest_entity_type = 'paragraph';
      $dest_bundle      = str_replace('field_', '', $dest_bundle);
    }

    if ($entity_type == 'paragraphs_item') {
      $dest_entity_type = 'paragraph';
    }

    if ($entity_type == 'menu_links') {
      $dest_entity_type = 'menu_link_content';
      $bundle_map       = [
        'main-menu' => 'main',
        'management' => 'admin',
        'navigation' => 'tools',
        'user-menu' => 'account',
      ];
      $dest_bundle      = $bundle_map[$bundle] ?? $bundle;
    }

    if ($entity_type == 'url_alias') {
      $dest_entity_type = $dest_bundle = 'path_alias';
    }

    if ($entity_type == 'block_custom') {
      $dest_entity_type = 'block_content';
      $dest_bundle      = 'basic';
    }

    if ($entity_type == 'file_private') {
      $dest_entity_type = 'file';
    }

    // Set the table name based on the entity type.
    $table_map = [
      'user' => 'migrate_map_d7_user',
      'node' => 'migrate_map_d7_node_complete__' . $bundle,
      'taxonomy_term' => 'migrate_map_d7_taxonomy_term__' . $bundle,
      'comment' => 'migrate_map_d7_comment',
      'field_collection_item' => 'migrate_map_d7_field_collection__' . $dest_bundle,
      'paragraphs_item' => 'migrate_map_d7_paragraphs__' . $bundle,
      'menu_links' => 'migrate_map_d7_menu_links',
      'url_alias' => 'migrate_map_d7_url_alias',
      'block_custom' => 'migrate_map_d7_custom_block',
      'file' => 'migrate_map_d7_file',
      'file_private' => 'migrate_map_d7_file_private',
    ];

    // Default generated table names, limited to 63 characters.
    $prefix_length = 0;

    if ($database = $this->state->get('data_migration_report_database')) {
      $prefix_length = strlen($database['prefix']);
    }

    $dest_map_table = mb_substr($table_map[$entity_type], 0, 63 - $prefix_length);

    return [
      'dest_entity_type' => $dest_entity_type,
      'dest_bundle' => $dest_bundle,
      'dest_map_table' => $dest_map_table,
    ];
  }

  /**
   * Returns the fields list of a given entity type bundle.
   *
   * @param string $entity_type
   *   The entity type.
   * @param string $bundle
   *   The entity bundle.
   *
   * @return array
   *   The fields list.
   */
  public function contentTypeFields(string $entity_type, string $bundle): array {
    $fields = [];

    if (!empty($entity_type) && !empty($bundle)) {
      // Query to fetch fields for Drupal 7.
      $d7_fields_query      = "SELECT fc.field_name, fc.type ";
      $d7_fields_query     .= "FROM {field_config} fc ";
      $d7_fields_query     .= "LEFT JOIN {field_config_instance} fci ON fci.field_id = fc.id ";
      $d7_fields_query     .= "WHERE fci.deleted = 0 AND fci.entity_type = :entity_type AND fci.bundle = :bundle ";
      $d7_fields_query     .= "ORDER BY fc.field_name ASC";
      $d7_fields_query_args = [
        ':entity_type' => $entity_type,
        ':bundle' => $bundle,
      ];
      $d7_fields            = $this->getValues($d7_fields_query, 'd7', 'two_column', $d7_fields_query_args);

      // Get destination entity mapping.
      $dest_mapping     = $this->getDestinationMapping($entity_type, $bundle);
      $dest_entity_type = $dest_mapping['dest_entity_type'];
      $dest_bundle      = $dest_mapping['dest_bundle'];

      // Initialize an array for Drupal 10 field information.
      $d10_fields            = [];
      $d10_field_definitions = $this->entityFieldManager->getFieldDefinitions($dest_entity_type, $dest_bundle);

      // Extract field names and types for Drupal 10.
      foreach ($d10_field_definitions as $field_definition) {
        if ($field_definition instanceof FieldConfigInterface) {
          $d10_fields[$field_definition->get('field_name')] = $field_definition->get('field_type');
        }
      }

      // Combine Drupal 7 and Drupal 10 field information into a single array.
      $fields = [
        'd7' => $d7_fields,
        'd10' => $d10_fields,
      ];
    }

    return $fields;
  }

  /**
   * Generates the content mapping yaml file.
   *
   * @param string $entity_type
   *   The entity type.
   * @param string $bundle
   *   The entity bundle.
   * @param array $fields
   *   The fields list.
   */
  public function generateMappingYaml(string $entity_type, string $bundle, array $fields = []): void {
    // Get destination entity mapping.
    $dest_mapping     = $this->getDestinationMapping($entity_type, $bundle);
    $dest_entity_type = $dest_mapping['dest_entity_type'];
    $dest_bundle      = $dest_mapping['dest_bundle'];
    $dest_map_table   = $dest_mapping['dest_map_table'];

    // Initialize the YAML content.
    $yaml_content = [
      'source' => [
        'entity_type' => $entity_type,
        'bundle' => $bundle,
      ],
      'table' => $dest_map_table,
      'fields' => [],
      'destination' => [
        'entity_type' => $dest_entity_type,
        'bundle' => $dest_bundle,
      ],
    ];

    // Loop through the fields and add them to the YAML content.
    if (!empty($fields)) {
      foreach ($fields['d7'] as $name => $type) {
        $i = count($yaml_content['fields']);

        $yaml_content['fields'][$i]['d7'] = [
          'name' => $name,
          'type' => $type,
        ];

        if (isset($fields['d10'][$name])) {
          $yaml_content['fields'][$i]['d10'] = [
            'name' => $name,
            'type' => $fields['d10'][$name],
          ];
        }
      }
    }

    // Convert the YAML content to YAML format.
    $yaml = Yaml::dump($yaml_content, 4, 2);

    // Prepare the directory and ensure it's writable.
    $directory = "public://data_migration_report/migration-mapping/$entity_type/";

    if (!$this->fileSystem->prepareDirectory($directory, FileSystemInterface::CREATE_DIRECTORY)) {
      $this->logger()->error($this->t('Unable to create directory @directory. The public directory is not writable.', ['@directory' => $directory]));
      return;
    }

    // Save the Migration mapping file.
    $file_path = $directory . $entity_type . '_' . $bundle . '.yml';
    $overwrite = TRUE;

    if (file_exists($file_path)) {
      $existing_data = $this->getParser()->parse(file_get_contents($file_path)) ?? [];

      if ($existing_data != $yaml_content) {
        $overwrite = $this->io()->confirm('Migration mapping YAML file ' . $file_path . ' already exists and has been modified. Do you want to overwrite it?');
      }
      else {
        $overwrite = FALSE;
        $this->logger()->notice($this->t('Migration mapping YAML file @file_path already exists with the same content, so it is being skipped.', [
          '@file_path' => $file_path,
        ]));
      }
    }

    if ($overwrite) {
      $mapping_file = $this->fileSystem->saveData($yaml, $file_path, FileExists::Replace);

      if ($mapping_file) {
        $path = $this->fileUrlGenerator->generateAbsoluteString($mapping_file);
        $this->logger()->log('success', $this->t('Migration mapping YAML file for @entity_type - @bundle generated at @path', [
          '@entity_type' => $entity_type,
          '@bundle' => $bundle,
          '@path' => $path,
        ]));
      }
      else {
        $this->logger()->error($this->t('An error occurred while creating the Migration mapping YAML file for @entity_type - @bundle at @path', [
          '@entity_type' => $entity_type,
          '@bundle' => $bundle,
          '@path' => $mapping_file,
        ]));
      }
    }
  }

  /**
   * Drush Command to run Migration testing.
   *
   * @param string $entity_type
   *   The type of the content entity being migrated.
   * @param string $bundle
   *   The bundle of the content entity.
   * @param array $options
   *   Array of options.
   *
   * @option integer limit
   *   Number of items to process.
   * @option array ids
   *   List of source content ids to process.
   *
   * @command migration:test
   * @usage drush mt node page --limit=10
   * @aliases mt
   */
  public function migrationTesting(
    string $entity_type,
    string $bundle,
    array $options = ['limit' => 0, 'ids' => []],
  ): void {
    // Load content data from the YML file.
    $path = "public://data_migration_report/migration-mapping/$entity_type/";
    $file = $path . $entity_type . '_' . $bundle . '.yml';

    if (!file_exists($file)) {
      $this->logger()->error($this->t('Migration mapping yml file not found at @path', [
        '@path' => $file,
      ]));
      return;
    }

    // Check source database connection and log an error if it's not found.
    $status = $this->checkSourceConnection();

    if (!$status) {
      $this->logger()->error($this->t('No source database connection found.'));
      $this->logger()->error($this->t('Please configure the source database credentials at :config_page', [
        ':config_page' => Url::fromRoute('data_migration_report.migrate_database_credentials', [], ['absolute' => TRUE])->toString(),
      ]));
      return;
    }

    $content_data     = $this->getParser()->parse(file_get_contents($file)) ?? [];
    $dest_entity_type = $content_data['destination']['entity_type'];
    $dest_bundle      = $content_data['destination']['bundle'];
    $table            = $content_data['table'];

    if (empty($content_data)) {
      $this->logger()->error($this->t('No content found in the YML file.'));
      return;
    }

    // Extract fields data from the content data.
    $errors = $fields = $fields['d7'] = $fields['d10'] = [];

    foreach ($content_data['fields'] as $field) {
      if (isset($field['d7'])) {
        $fields['d7'][] = $field['d7'];
      }
      if (isset($field['d10'])) {
        $fields['d10'][] = $field['d10'];
      }
      if (isset($field['d7']) && isset($field['d10'])) {
        if ($field['d7']['name'] != $field['d10']['name']) {
          $this->fieldNameMapping[$field['d10']['name']] = $field['d7']['name'];
        }
      }
    }

    // Prepare SQL queries for fetching Drupal 7 IDs.
    $d7_ids_query_args = [];
    $ids_condition     = '';

    if ($options['ids']) {
      if ($entity_type == 'user') {
        $ids_condition = "AND uid IN (:ids[])";
      }
      elseif ($entity_type == 'node') {
        $ids_condition = "AND nid IN (:ids[])";
      }
      elseif ($entity_type == 'taxonomy_term') {
        $ids_condition = "AND t.tid IN (:ids[])";
      }
      elseif ($entity_type == 'block_custom') {
        $ids_condition = "AND bid IN (:ids[])";
      }
      elseif ($entity_type == 'comment') {
        $ids_condition = "AND c.cid IN (:ids[])";
      }
      elseif ($entity_type == 'menu_links') {
        $ids_condition = "AND mlid IN (:ids[])";
      }
      elseif ($entity_type == 'url_alias') {
        $ids_condition = "WHERE pid IN (:ids[])";
      }
      elseif ($entity_type == 'file' || $entity_type == 'file_private') {
        $ids_condition = "AND fid IN (:ids[])";
      }
      elseif ($entity_type == 'field_collection_item' || $entity_type == 'paragraphs_item') {
        $ids_condition = "AND item_id IN (:ids[])";
      }
      $d7_ids_query_args[':ids[]'] = (array) $options['ids'];
    }

    if ($entity_type == 'user') {
      $d7_ids_query = "SELECT uid FROM {users} WHERE uid > 0 $ids_condition ORDER BY uid";
    }
    elseif ($entity_type == 'node') {
      $d7_ids_query                 = "SELECT nid FROM {node} WHERE type = :bundle $ids_condition ORDER BY nid";
      $d7_ids_query_args[':bundle'] = $bundle;
    }
    elseif ($entity_type == 'taxonomy_term') {
      $d7_ids_query                 = "SELECT t.tid FROM {taxonomy_term_data} t LEFT JOIN {taxonomy_vocabulary} tv ON tv.vid = t.vid WHERE tv.machine_name = :bundle $ids_condition ORDER BY t.tid";
      $d7_ids_query_args[':bundle'] = $bundle;
    }
    elseif ($entity_type == 'block_custom') {
      $d7_ids_query = "SELECT bid FROM {block_custom} $ids_condition ORDER BY bid";
    }
    elseif ($entity_type == 'comment') {
      $d7_ids_query                 = "SELECT c.cid FROM {comment} c LEFT JOIN {node} n ON n.nid = c.nid WHERE n.type = :bundle $ids_condition ORDER BY c.cid";
      $d7_ids_query_args[':bundle'] = str_replace("comment_node_", "", $bundle);
    }
    elseif ($entity_type == 'menu_links') {
      $d7_ids_query                 = "SELECT mlid FROM {menu_links} WHERE menu_name = :bundle $ids_condition ORDER BY mlid";
      $d7_ids_query_args[':bundle'] = $bundle;
    }
    elseif ($entity_type == 'url_alias') {
      $d7_ids_query = "SELECT pid FROM {url_alias} $ids_condition ORDER BY pid";
    }
    elseif ($entity_type == 'file') {
      $d7_ids_query = "SELECT fid FROM {file_managed} WHERE uri LIKE 'public://%' $ids_condition ORDER BY fid";
    }
    elseif ($entity_type == 'file_private') {
      $d7_ids_query = "SELECT fid FROM {file_managed} WHERE uri LIKE 'private://%' $ids_condition ORDER BY fid";
    }
    elseif ($entity_type == 'field_collection_item') {
      $d7_ids_query                 = "SELECT item_id FROM {field_collection_item} WHERE field_name = :bundle $ids_condition ORDER BY item_id";
      $d7_ids_query_args[':bundle'] = $bundle;
    }
    elseif ($entity_type == 'paragraphs_item') {
      $d7_ids_query                 = "SELECT item_id FROM {paragraphs_item} WHERE bundle = :bundle $ids_condition ORDER BY item_id";
      $d7_ids_query_args[':bundle'] = $bundle;
    }

    if (!empty($d7_ids_query)) {
      if ($options['limit'] && $options['limit'] > 0) {
        $d7_ids_query = $d7_ids_query . " LIMIT " . $options['limit'];
      }

      $d7_ids = $this->getValues($d7_ids_query, 'd7', 'one_column', $d7_ids_query_args);
    }
    else {
      $d7_ids_query = $this->database->select($content_data['table'], 'n')
        ->fields('n', ['sourceid1'])
        ->orderBy('sourceid1');

      if ($options['limit'] && $options['limit'] > 0) {
        $d7_ids_query->range(0, $options['limit']);
      }

      $d7_ids = $d7_ids_query->execute()->fetchCol();
    }

    if (empty($d7_ids)) {
      $this->logger()->notice($this->t('No data found for @entity_type - @bundle in Drupal 7.', [
        '@entity_type' => $entity_type,
        '@bundle' => $bundle,
      ]));
      return;
    }

    // Load content mapping.
    $content_mapping_query = $this->database->select($content_data['table'], 'n')
      ->fields('n', ['sourceid1', 'destid1'])
      ->orderBy('sourceid1');
    $content_mapping = $content_mapping_query->execute()->fetchAllKeyed();

    // Load destination IDs.
    $d10_ids_query = $this->database->select($content_data['table'], 'n')
      ->fields('n', ['destid1'])
      ->condition('destid1', NULL, 'IS NOT')
      ->orderBy('sourceid1');

    if ($options['ids']) {
      $dest_ids = [];

      foreach ($options['ids'] as $source_id) {
        if (array_key_exists($source_id, $content_mapping)) {
          $dest_ids[] = $content_mapping[$source_id];
        }
      }

      if (!empty($dest_ids)) {
        $d10_ids_query->condition('destid1', $dest_ids, 'IN');
      }
    }

    if ($options['limit'] && $options['limit'] > 0) {
      $d10_ids_query->range(0, $options['limit']);
    }

    $d10_ids = $d10_ids_query->execute()->fetchCol();

    // Load source and destination content.
    $d7_contents  = $this->getEntityValues($entity_type, $content_data['fields'], 'd7', $d7_ids);
    $d10_contents = $this->getEntityValues($dest_entity_type, $content_data['fields'], 'd10', $d10_ids);

    // Preprocess the content.
    $processed_data = $this->executePreprocess($entity_type, $dest_entity_type, $fields, $d7_contents, $d10_contents);
    $d7_contents    = $processed_data['source'];
    $d10_contents   = $processed_data['dest'];

    // Get the column names of the source and destination content.
    $column_d7  = array_keys(reset($d7_contents));
    $column_d10 = array_keys(reset($d10_contents));

    // Remove the ID column from the comparison, as the ID value will not
    // always be the same after migration.
    $id_column = '';

    if ($entity_type == 'user') {
      $id_column = 'uid';
    }
    elseif ($entity_type == 'node') {
      $id_column = 'nid';
    }
    elseif ($entity_type == 'taxonomy_term') {
      $id_column = 'tid';
    }
    elseif ($entity_type == 'block_custom') {
      $id_column = 'bid';
    }
    elseif ($entity_type == 'comment') {
      $id_column = 'cid';
    }
    elseif ($entity_type == 'menu_links') {
      $id_column = 'mlid';
    }
    elseif ($entity_type == 'url_alias') {
      $id_column = 'pid';
    }
    elseif ($entity_type == 'file' || $entity_type == 'file_private') {
      $id_column = 'fid';
    }
    elseif ($entity_type == 'field_collection_item' || $entity_type == 'paragraphs_item') {
      $id_column = 'item_id';
    }

    if (!empty($id_column)) {
      unset($column_d7[array_search($id_column, $column_d7)]);
      // To restore array key sequence.
      $column_d7 = array_values($column_d7);
    }

    $installed_drupal_version = intval(\Drupal::VERSION);
    $destination_value        = "d{$installed_drupal_version}_value";

    // Find errors in the content.
    foreach ($d7_contents as $key => $entity) {
      $source_id = $key;
      $dest_id   = $content_mapping[$source_id] ?? NULL;
      $dest_flag = TRUE;

      if ($dest_id) {
        if (array_key_exists($dest_id, $d10_contents)) {
          for ($i = 0; $i < count($column_d7); $i++) {
            $d7_column      = $column_d7[$i];
            $d7_field       = $this->fieldColumnMapping['d7'][$d7_column] ?? 'default';
            $d10_column_key = $this->fieldPreprocessMapping[$d7_column] ?? $d7_column;
            $d10_column_key = $this->fieldNamePreprocessMapping[$d10_column_key] ?? $d10_column_key;

            if (in_array($d7_column, $column_d10)) {
              $d10_column = $d7_column;

              if (is_array($entity[$d7_column]) && is_array($d10_contents[$dest_id][$d10_column])) {
                if (array_diff($entity[$d7_column], $d10_contents[$dest_id][$d10_column])) {
                  $errors[$source_id][$d7_field][] = [
                    'd7_value' => [
                      $d7_column => $entity[$d7_column],
                    ],
                    $destination_value => [
                      $d10_column_key => $d10_contents[$dest_id][$d10_column],
                    ],
                  ];
                }
              }
              else {
                if ($entity[$d7_column] != $d10_contents[$dest_id][$d10_column]) {
                  $errors[$source_id][$d7_field][] = [
                    'd7_value' => [
                      $d7_column => $entity[$d7_column],
                    ],
                    $destination_value => [
                      $d10_column_key => $d10_contents[$dest_id][$d10_column],
                    ],
                  ];
                }
              }
            }
            else {
              if (!empty($entity[$d7_column])) {
                $errors[$source_id][$d7_field][] = [
                  'd7_value' => [
                    $d7_column => $entity[$d7_column],
                  ],
                  $destination_value => [
                    $d10_column_key => '',
                  ],
                ];
              }
            }
          }
        }
        else {
          $dest_flag = FALSE;
        }
      }
      else {
        $dest_flag = FALSE;
      }

      if (!$dest_flag) {
        for ($i = 0; $i < count($column_d7); $i++) {
          $d7_column      = $column_d7[$i];
          $d7_field       = $this->fieldColumnMapping['d7'][$d7_column] ?? 'default';
          $d10_column_key = $this->fieldPreprocessMapping[$d7_column] ?? $d7_column;
          $d10_column_key = $this->fieldNamePreprocessMapping[$d10_column_key] ?? $d10_column_key;

          if (!empty($entity[$d7_column])) {
            $errors[$source_id][$d7_field][] = [
              'd7_value' => [
                $d7_column => $entity[$d7_column],
              ],
              $destination_value => [
                $d10_column_key => '',
              ],
            ];
          }
        }
      }
    }

    $total  = count($d7_ids);
    $failed = count($errors);
    $passed = $total - $failed;

    $report_markup = [
      'entity_type' => $entity_type,
      'bundle' => $bundle,
      'dest_entity_type' => $dest_entity_type,
      'dest_bundle' => $dest_bundle,
      'table' => $table,
      'total' => $total,
      'passed' => $passed,
      'failed' => $failed,
      'errors' => $errors,
      'pass' => round(($passed / $total) * 100, 2),
      'fail' => round(($failed / $total) * 100, 2),
    ];

    $report_markup = [
      '#theme' => 'migration_report',
      '#data' => $report_markup,
    ];

    $report_markup = $this->renderer->renderInIsolation($report_markup);

    // Print the summary table.
    $output = new ConsoleOutput();
    $output->writeln('<info>Migration Summary</info>');

    if ($entity_type == 'user' || $entity_type == 'url_alias' || $entity_type == 'block_custom' || $entity_type == 'file' || $entity_type == 'file_private') {
      $summary_header = ['Type', 'Total', 'Passed', 'Failed'];
    }
    else {
      $type = ucwords(str_replace('_', ' ', $entity_type));
      $summary_header = [$type, 'Total', 'Passed', 'Failed'];
    }

    $summary_rows = [
      $bundle => [
        'type' => $bundle,
        'total' => $total,
        'passed' => $passed,
        'failed' => $failed,
      ],
    ];
    $this->printTable($summary_header, $summary_rows, $output);

    // HTML Report Generation.
    $directory = "public://data_migration_report/migration-reports/$entity_type/";

    // Prepare the directory and ensure it's writable.
    if (!$this->fileSystem->prepareDirectory($directory, FileSystemInterface::CREATE_DIRECTORY)) {
      $this->logger()->error($this->t('Unable to create directory @directory. The public directory is not writable.', ['@directory' => $directory]));
    }

    // Save the Migration Summary Report file.
    $report_file_path = $directory . 'migration_report-' . $entity_type . '-' . $bundle . '.html';
    $report_file      = $this->fileSystem->saveData($report_markup->__toString(), $report_file_path, FileExists::Replace);

    if ($report_file) {
      $path = $this->fileUrlGenerator->generateAbsoluteString($report_file);
      $this->logger()->log('success', $this->t('Migration Report generated at @path', ['@path' => $path]));
    }
    else {
      $this->logger()->error($this->t('An error occurred while creating the Migration Summary Report HTML file.'));
    }
  }

  /**
   * Check the source database connection.
   *
   * @return bool
   *   TRUE if the source database connection is available, FALSE otherwise.
   */
  public function checkSourceConnection(): bool {
    $status = FALSE;

    // Check if the "migrate" connection is available.
    if (Database::getConnection('default', 'migrate')) {
      $status                         = TRUE;
      $this->sourceDatabaseConnection = Database::getConnection('default', 'migrate');
    }
    else {
      // Get the driver and database values from the state.
      $driver   = $this->state->get('data_migration_report_database_driver');
      $database = $this->state->get('data_migration_report_database');

      // If both values are present, create the source database connection.
      if (!empty($driver) && !empty($database)) {
        // Get the namespace and driver from the database types.
        $drivers           = CredentialForm::getDatabaseTypes();
        $reflection        = new \ReflectionClass($drivers[$driver]);
        $install_namespace = $reflection->getNamespaceName();
        // Cut the trailing \Install from namespace.
        $database['namespace']          = substr($install_namespace, 0, strrpos($install_namespace, '\\'));
        $database['driver']             = $driver;
        $this->sourceDatabaseConnection = $this->getConnection($database);
        $status                         = TRUE;
      }
    }

    return $status;
  }

  /**
   * Prints tabular data.
   *
   * @param array $header
   *   The header for the table.
   * @param array $rows
   *   The rows of data to be displayed in the table.
   * @param \Symfony\Component\Console\Output\ConsoleOutput $output
   *   The console output to be used for rendering the table.
   */
  public function printTable(array $header, array $rows, ConsoleOutput $output): void {
    // Initialize table.
    $table = new Table($output);

    // Set the table header.
    $table->setHeaders($header);
    // Add rows to the table.
    $table->addRows($rows);

    // Render the table.
    $table->render();
  }

  /**
   * Get the YAML parser service.
   *
   * @return \Symfony\Component\Yaml\Parser
   *   The YAML parser service.
   */
  protected function getParser(): Parser {
    // Create a new YAML parser if it has not been created yet.
    if (!isset($this->parser)) {
      $this->parser = new Parser();
    }

    // Return the YAML parser service.
    return $this->parser;
  }

  /**
   * Get the values of the entity fields from the database.
   *
   * @param string $entity_type
   *   The type of entity.
   * @param array $fields
   *   The fields of the entity.
   * @param string $drupal_version
   *   The version of Drupal.
   * @param array $ids
   *   The IDs of the entities.
   *
   * @return array
   *   The values of the entity fields.
   */
  public function getEntityValues(string $entity_type, array $fields, string $drupal_version, array $ids): array {
    // Initialize the base columns array and the query.
    $base_columns = [];
    $entity_table = $prefix = $query = '';

    // Set prefix and entity table based on the entity type and Drupal version.
    if ($entity_type == 'user') {
      $prefix       = $drupal_version == 'd10' ? 'user__' : 'field_data_';
      $entity_table = $drupal_version == 'd10' ? '{users_field_data}' : '{users}';
      $base_columns = [
        $entity_table . '.uid',
        $entity_table . '.name',
        $entity_table . '.pass',
        $entity_table . '.mail',
        $entity_table . '.status',
        $entity_table . '.timezone',
      ];
    }
    elseif ($entity_type == 'node') {
      $prefix         = $drupal_version == 'd10' ? 'node__' : 'field_data_';
      $entity_table   = $drupal_version == 'd10' ? '{node_field_data}' : '{node}';
      $base_columns   = [
        $entity_table . '.nid',
        $entity_table . '.title',
        $entity_table . '.type',
        $entity_table . '.status',
      ];
      $base_columns[] = $drupal_version == 'd10' ? $entity_table . '.langcode' : $entity_table . '.language';
    }
    elseif ($entity_type == 'taxonomy_term') {
      $prefix         = $drupal_version == 'd10' ? 'taxonomy_term__' : 'field_data_';
      $entity_table   = $drupal_version == 'd10' ? '{taxonomy_term_field_data}' : '{taxonomy_term_data}';
      $base_columns   = [
        $entity_table . '.tid',
        $entity_table . '.vid',
        $entity_table . '.name',
      ];
      $base_columns[] = $drupal_version == 'd10' ? $entity_table . '.description__value' : $entity_table . '.description';
      $base_columns[] = $drupal_version == 'd10' ? $entity_table . '.description__format' : $entity_table . '.format';
    }
    elseif ($entity_type == 'block_custom' || $entity_type == 'block_content') {
      $entity_table   = $drupal_version == 'd10' ? '{block_content_field_data}' : '{block_custom}';
      $base_columns   = [
        $entity_table . '.info',
      ];
      $base_columns[] = $drupal_version == 'd10' ? $entity_table . '.id' : $entity_table . '.bid';
      $base_columns[] = $drupal_version == 'd7' ? $entity_table . '.body' : '';
      $base_columns[] = $drupal_version == 'd7' ? $entity_table . '.format' : '';
    }
    elseif ($entity_type == 'comment') {
      $prefix         = $drupal_version == 'd10' ? 'comment__' : 'field_data_';
      $entity_table   = $drupal_version == 'd10' ? '{comment_field_data}' : '{comment}';
      $base_columns   = [
        $entity_table . '.cid',
        $entity_table . '.pid',
        $entity_table . '.uid',
        $entity_table . '.subject',
        $entity_table . '.hostname',
        $entity_table . '.created',
        $entity_table . '.changed',
        $entity_table . '.status',
        $entity_table . '.thread',
        $entity_table . '.name',
        $entity_table . '.mail',
        $entity_table . '.homepage',
      ];
      $base_columns[] = $drupal_version == 'd10' ? $entity_table . '.langcode' : $entity_table . '.language';
      $base_columns[] = $drupal_version == 'd10' ? $entity_table . '.entity_id' : $entity_table . '.nid';
    }
    elseif ($entity_type == 'menu_links' || $entity_type == 'menu_link_content') {
      $entity_table   = $drupal_version == 'd10' ? '{menu_link_content_data}' : '{menu_links}';
      $base_columns   = [
        $entity_table . '.external',
        $entity_table . '.expanded',
        $entity_table . '.weight',
      ];
      $base_columns[] = $drupal_version == 'd10' ? $entity_table . '.id' : $entity_table . '.mlid';
      $base_columns[] = $drupal_version == 'd7' ? $entity_table . '.plid' : '';
      $base_columns[] = $drupal_version == 'd10' ? $entity_table . '.link__uri' : $entity_table . '.link_path';
      $base_columns[] = $drupal_version == 'd10' ? $entity_table . '.title' : $entity_table . '.link_title';
      $base_columns[] = $drupal_version == 'd10' ? $entity_table . '.link__options' : $entity_table . '.options';
    }
    elseif ($entity_type == 'url_alias' || $entity_type == 'path_alias') {
      $entity_table   = $drupal_version == 'd10' ? '{path_alias}' : '{url_alias}';
      $base_columns   = [
        $entity_table . '.alias',
      ];
      $base_columns[] = $drupal_version == 'd10' ? $entity_table . '.id' : $entity_table . '.pid';
      $base_columns[] = $drupal_version == 'd10' ? $entity_table . '.path' : $entity_table . '.source';
      $base_columns[] = $drupal_version == 'd10' ? $entity_table . '.langcode' : $entity_table . '.language';
    }
    elseif ($entity_type == 'file' || $entity_type == 'file_private') {
      $entity_table = 'file_managed';
      $base_columns = [
        $entity_table . '.fid',
        $entity_table . '.filename',
        $entity_table . '.uri',
        $entity_table . '.filemime',
        $entity_table . '.status',
      ];
    }
    elseif ($entity_type == 'field_collection_item') {
      $prefix         = $drupal_version == 'd10' ? 'paragraph__' : 'field_data_';
      $entity_table   = $drupal_version == 'd10' ? '{paragraphs_item_field_data}' : '{field_collection_item}';
      $base_columns[] = $drupal_version == 'd10' ? $entity_table . '.id' : $entity_table . '.item_id';
    }
    elseif ($entity_type == 'paragraphs_item' || $entity_type == 'paragraph') {
      $prefix         = $drupal_version == 'd10' ? 'paragraph__' : 'field_data_';
      $entity_table   = $drupal_version == 'd10' ? '{paragraphs_item_field_data}' : '{paragraphs_item}';
      $base_columns[] = $drupal_version == 'd10' ? $entity_table . '.id' : $entity_table . '.item_id';
    }

    // Join the base columns with the field columns.
    $columns = implode(",", array_filter($base_columns));
    $tables  = $entity_table;

    // Join tables with the field tables and add the field columns to the query.
    if (!empty($fields)) {
      foreach ($fields as $field) {
        if (isset($field[$drupal_version]['name']) && $field[$drupal_version]['type']) {
          $field_name    = $field[$drupal_version]['name'];
          $field_type    = $field[$drupal_version]['type'];
          $field_columns = $this->getFieldColumnList($field_name, $field_type, $prefix, $drupal_version);

          if (!empty($field_columns)) {
            $columns .= !empty($columns) ? "," . $field_columns : $field_columns;

            if ($entity_type == 'user') {
              $tables .= " LEFT JOIN {$prefix}{$field_name} ON {$prefix}{$field_name}.entity_id = {$entity_table}.uid";
            }
            elseif ($entity_type == 'node') {
              $tables .= " LEFT JOIN {$prefix}{$field_name} ON {$prefix}{$field_name}.entity_id = {$entity_table}.nid";
            }
            elseif ($entity_type == 'taxonomy_term') {
              $tables .= " LEFT JOIN {$prefix}{$field_name} ON {$prefix}{$field_name}.entity_id = {$entity_table}.tid";
            }
            elseif ($entity_type == 'comment') {
              $tables .= " LEFT JOIN {$prefix}{$field_name} ON {$prefix}{$field_name}.entity_id = {$entity_table}.cid";
            }
            elseif ($entity_type == 'field_collection_item' ||$entity_type == 'paragraphs_item') {
              $tables .= " LEFT JOIN {$prefix}{$field_name} ON {$prefix}{$field_name}.entity_id = {$entity_table}.item_id";
            }
            elseif ($entity_type == 'paragraph') {
              $tables .= " LEFT JOIN {$prefix}{$field_name} ON {$prefix}{$field_name}.entity_id = {$entity_table}.id";
            }

            if ($drupal_version == 'd7') {
              $tables .= " AND {$prefix}{$field_name}.entity_type = $entity_type";
            }
          }
        }
      }
    }

    if ($entity_type == 'user') {
      if ($drupal_version == 'd10') {
        $columns .= ",{user__roles}.roles_target_id";
        $tables  .= " LEFT JOIN {user__roles} ON {user__roles}.entity_id = {$entity_table}.uid";
      }
      else {
        $columns .= ",{users_roles}.rid";
        $tables  .= " LEFT JOIN {users_roles} ON {users_roles}.uid = {$entity_table}.uid";
      }
    }

    if ($entity_type == 'menu_link_content') {
      $columns .= ",{menu_link_content}.id parent";
      $tables  .= " LEFT JOIN {menu_link_content} ON {$entity_table}.parent = CONCAT('menu_link_content:', {menu_link_content}.uuid)";
    }

    if ($entity_type == 'block_content') {
      $columns .= ",{block_content__body}.body_value,{block_content__body}.body_format";
      $tables  .= " LEFT JOIN {block_content__body} ON {$entity_table}.id = {block_content__body}.entity_id";
    }

    // Build the query and execute it to get the values.
    if ($entity_type == 'user') {
      $query = "SELECT $columns FROM $tables WHERE {$entity_table}.uid IN (:ids[]) ORDER BY {$entity_table}.uid ASC";
    }
    elseif ($entity_type == 'node') {
      $query = "SELECT $columns FROM $tables WHERE {$entity_table}.nid IN (:ids[]) ORDER BY {$entity_table}.nid ASC";
    }
    elseif ($entity_type == 'taxonomy_term') {
      $query = "SELECT $columns FROM $tables WHERE {$entity_table}.tid IN (:ids[]) ORDER BY {$entity_table}.tid ASC";
    }
    elseif ($entity_type == 'block_custom') {
      $query = "SELECT $columns FROM $tables WHERE {$entity_table}.bid IN (:ids[]) ORDER BY {$entity_table}.bid ASC";
    }
    elseif ($entity_type == 'comment') {
      $query = "SELECT $columns FROM $tables WHERE {$entity_table}.cid IN (:ids[]) ORDER BY {$entity_table}.cid ASC";
    }
    elseif ($entity_type == 'menu_links') {
      $query = "SELECT $columns FROM $tables WHERE {$entity_table}.mlid IN (:ids[]) ORDER BY {$entity_table}.mlid ASC";
    }
    elseif ($entity_type == 'url_alias') {
      $query = "SELECT $columns FROM $tables WHERE {$entity_table}.pid IN (:ids[]) ORDER BY {$entity_table}.pid ASC";
    }
    elseif ($entity_type == 'file' || $entity_type == 'file_private') {
      $query = "SELECT $columns FROM $tables WHERE {$entity_table}.fid IN (:ids[]) ORDER BY {$entity_table}.fid ASC";
    }
    elseif ($entity_type == 'field_collection_item' || $entity_type == 'paragraphs_item') {
      $query = "SELECT $columns FROM $tables WHERE {$entity_table}.item_id IN (:ids[]) ORDER BY {$entity_table}.item_id ASC";
    }
    elseif ($entity_type == 'menu_link_content' || $entity_type == 'path_alias' || $entity_type == 'paragraph' || $entity_type == 'block_content') {
      $query = "SELECT $columns FROM $tables WHERE {$entity_table}.id IN (:ids[]) ORDER BY {$entity_table}.id ASC";
    }

    $query_args = [
      ':ids[]' => $ids,
    ];

    return $this->getValues($query, $drupal_version, $entity_type, $query_args);
  }

  /**
   * Get the field columns for a given field name and field type.
   *
   * @param string $field_name
   *   The name of the field.
   * @param string $field_type
   *   The type of the field.
   * @param string $prefix
   *   The prefix to use for the field tables.
   * @param string $drupal_version
   *   The Drupal version.
   *
   * @return string
   *   The list of field columns.
   */
  public function getFieldColumnList(string $field_name, string $field_type, string $prefix, string $drupal_version): string {
    // Initialize variables.
    $columns       = [];
    $field_columns = "";

    // Get the field type mapping.
    $field_type_data    = $this->getFieldTypeColumnMapping();
    $field_type_mapping = $field_type_data[$drupal_version][$field_type] ?? [];

    foreach ($field_type_mapping as $column) {
      // Generate the field columns.
      $column    = str_replace("{field_name}", $field_name, $column);
      $columns[] = "{$prefix}{$field_name}.$column";

      // Store the field column mapping.
      $this->fieldColumnMapping[$drupal_version][$column] = $field_name;
    }

    // Combine the field columns into a string.
    if (!empty($columns)) {
      $field_columns = implode(",", $columns);
    }

    return $field_columns;
  }

  /**
   * Get the field type columns mapping data.
   *
   * @return array
   *   The array of field type columns mapping.
   */
  public function getFieldTypeColumnMapping(): array {
    $path = $this->pathResolver->getPath('module', 'data_migration_report');
    $file = $path . '/field_types.yml';
    return $this->getParser()->parse(file_get_contents($file)) ?? [];
  }

  /**
   * Retrieve values from the database based on the given query.
   *
   * @param string $query
   *   The SQL query to execute.
   * @param string $drupal_version
   *   The Drupal version.
   * @param string $result_type
   *   The type of result to fetch.
   * @param array $arguments
   *   The arguments to pass to the query.
   *
   * @return array
   *   An associative array containing the fetched values.
   */
  public function getValues(string $query, string $drupal_version, string $result_type, array $arguments = []): array {
    // Determine the database connection based on the Drupal version.
    if ($drupal_version == 'd10') {
      $connection = $this->database;
    }

    if ($drupal_version == 'd7') {
      $connection = $this->sourceDatabaseConnection;
    }

    // If a database connection exists, execute the query and fetch the values.
    if (!empty($connection) && !empty($query)) {
      $result = $connection->query($query, $arguments);

      if ($result_type == 'user') {
        $result = $result->fetchAll();
        return $this->prepareAssociativeArray($result, 'uid');
      }
      elseif ($result_type == 'node') {
        $result = $result->fetchAll();
        return $this->prepareAssociativeArray($result, 'nid');
      }
      elseif ($result_type == 'taxonomy_term') {
        $result = $result->fetchAll();
        return $this->prepareAssociativeArray($result, 'tid');
      }
      elseif ($result_type == 'block_custom') {
        $result = $result->fetchAll();
        return $this->prepareAssociativeArray($result, 'bid');
      }
      elseif ($result_type == 'comment') {
        $result = $result->fetchAll();
        return $this->prepareAssociativeArray($result, 'cid');
      }
      elseif ($result_type == 'menu_links') {
        $result = $result->fetchAll();
        return $this->prepareAssociativeArray($result, 'mlid');
      }
      elseif ($result_type == 'url_alias') {
        $result = $result->fetchAll();
        return $this->prepareAssociativeArray($result, 'pid');
      }
      elseif ($result_type == 'file' || $result_type == 'file_private') {
        $result = $result->fetchAll();
        return $this->prepareAssociativeArray($result, 'fid');
      }
      elseif ($result_type == 'field_collection_item' ||$result_type == 'paragraphs_item') {
        $result = $result->fetchAll();
        return $this->prepareAssociativeArray($result, 'item_id');
      }
      elseif ($result_type == 'menu_link_content' || $result_type == 'path_alias' || $result_type == 'paragraph' || $result_type == 'block_content') {
        $result = $result->fetchAll();
        return $this->prepareAssociativeArray($result, 'id');
      }
      elseif ($result_type == 'one_column') {
        return $result->fetchAllKeyed(0, 0);
      }
      elseif ($result_type == 'two_column') {
        return $result->fetchAllKeyed();
      }
    }

    // If no values were fetched, return an empty array.
    return [];
  }

  /**
   * Prepare an associative array for provided data keyed by index column.
   *
   * @param array $query_result
   *   The query result array.
   * @param string $index_column
   *   The index column key.
   *
   * @return array
   *   An associative array containing the values.
   */
  public function prepareAssociativeArray(array $query_result, string $index_column): array {
    $data = [];

    if (!empty($query_result)) {
      foreach ($query_result as $row) {
        $id = $row->$index_column;

        foreach ($row as $key => $value) {
          if (isset($data[$id][$key])) {
            if (is_array($data[$id][$key])) {
              if (!in_array($value, $data[$id][$key])) {
                $data[$id][$key][] = $value;
              }
            }
            else {
              if ($data[$id][$key] != $value) {
                $data[$id][$key] = [$data[$id][$key], $value];
              }
            }
          }
          else {
            $data[$id][$key] = $value;
          }
        }
      }
    }

    return $data;
  }

  /**
   * Preprocess the data before comparing.
   *
   * @param string $entity_type
   *   The source entity type.
   * @param string $dest_entity_type
   *   The destination entity type.
   * @param array $fields
   *   An array of fields.
   * @param array $source
   *   The source data.
   * @param array $dest
   *   The destination data.
   *
   * @return array
   *   An array containing the preprocessed source and destination data.
   */
  public function executePreprocess(string $entity_type, string $dest_entity_type, array $fields, array $source, array $dest): array {
    // Set the source and destination data.
    $this->source = $source;
    $this->dest   = $dest;

    foreach ($fields as $drupal_version => $data) {
      foreach ($data as $field) {
        // Preprocessors field name mapping.
        if ($drupal_version == 'd10' && !empty($this->fieldNameMapping) && isset($this->fieldNameMapping[$field['name']])) {
          call_user_func([self::class, 'preprocessFieldName'], $field);
        }

        // Preprocessors field types.
        $method = str_replace("_", "", lcfirst(ucwords($field['type'], " /_")));
        if (method_exists(self::class, $method)) {
          call_user_func([self::class, $method], $field);
        }

        // Preprocessors field specific.
        $method = str_replace("_", "", lcfirst(ucwords($field['name'], " /_")));
        if (method_exists(self::class, $method)) {
          call_user_func([self::class, $method], $field);
        }
      }
    }

    // Preprocessors based on the main entity type.
    $method = str_replace("_", "", lcfirst(ucwords('entity_' . $entity_type, " /_")));
    if (method_exists(self::class, $method)) {
      call_user_func([self::class, $method], NULL);
    }

    if ($entity_type != $dest_entity_type) {
      $method = str_replace("_", "", lcfirst(ucwords('entity_' . $dest_entity_type, " /_")));
      if (method_exists(self::class, $method)) {
        call_user_func([self::class, $method], NULL);
      }
    }

    // Return the preprocessed source and destination data.
    return ['source' => $this->source, 'dest' => $this->dest];
  }

  /**
   * Preprocessor for different field name mapping.
   *
   * @param array $field
   *   The field definition containing the name of the field.
   */
  public function preprocessFieldName(array $field): void {
    // Get the field name.
    $field_name             = $field['name'];
    $field_type_column_data = $this->getFieldTypeColumnMapping();
    $field_columns          = $field_type_column_data['d10'][$field['type']];

    // Iterate over each record in the destination data.
    foreach ($this->dest as $key => $value) {
      foreach ($field_columns as $column) {
        $column = str_replace("{field_name}", $field_name, $column);

        if (array_key_exists($column, $this->dest[$key])) {
          $new_column = str_replace($field_name, $this->fieldNameMapping[$field_name], $column);
          $this->dest[$key][$new_column] = $this->dest[$key][$column];
          unset($this->dest[$key][$column]);

          // Keep track of the field renaming for later use.
          $this->fieldNamePreprocessMapping[$new_column] = $column;
        }
      }
    }
  }

  /**
   * Preprocessor for email field type.
   *
   * @param array $field
   *   The field definition containing the name of the email field.
   */
  public function email(array $field): void {
    // Get the field name.
    $field_name = $field['name'];

    // Iterate over each record in the destination data.
    foreach ($this->dest as $key => $value) {
      // If the record has a field named {field_name}_value, rename it to
      // {field_name}_email and remove the {field_name}_value field.
      if (array_key_exists($field_name . '_value', $this->dest[$key])) {
        $this->dest[$key][$field_name . '_email'] = $this->dest[$key][$field_name . '_value'];
        unset($this->dest[$key][$field_name . '_value']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping[$field_name . '_email'] = $field_name . '_value';
      }
    }
  }

  /**
   * Preprocessor for text_with_summary field type.
   *
   * This function removes carriage return, newline and tab characters from the
   * source and destination values of the field.
   *
   * @param array $field
   *   The field definition.
   */
  public function textWithSummary(array $field): void {
    $this->preprocessText($field);
  }

  /**
   * Preprocessor for text_long field type.
   *
   * This function removes carriage return, newline and tab characters from the
   * source and destination values of the field.
   *
   * @param array $field
   *   The field definition.
   */
  public function textLong(array $field): void {
    $this->preprocessText($field);
  }

  /**
   * Preprocessor for text fields.
   *
   * This function removes carriage return, newline and tab characters from the
   * source and destination values of the field.
   *
   * @param array $field
   *   The field definition array.
   */
  public function preprocessText(array $field): void {
    // Get the field name.
    $field_name = $field['name'];

    // Remove carriage return, newline & tab characters from source value.
    foreach ($this->source as $key => $value) {
      if (array_key_exists($field_name . '_value', $value) && !empty($value[$field_name . '_value'])) {
        $this->source[$key][$field_name . '_value'] = preg_replace('/[\r\n\t]/', '', $value[$field_name . '_value']);
      }
    }

    // Remove carriage return, newline & tab characters from destination value.
    foreach ($this->dest as $key => $value) {
      if (array_key_exists($field_name . '_value', $value) && !empty($value[$field_name . '_value'])) {
        $this->dest[$key][$field_name . '_value'] = preg_replace('/[\r\n\t]/', '', $value[$field_name . '_value']);
      }
    }
  }

  /**
   * Preprocessor for datetime field type.
   *
   * @param array $field
   *   The field information.
   */
  public function datetime(array $field): void {
    // Get the field name.
    $field_name = $field['name'];

    // Iterate over the destination array.
    foreach ($this->dest as $key => $value) {
      // Check if the field value exists and is not empty.
      if (array_key_exists($field_name . '_value', $value) && !empty($value[$field_name . '_value'])) {
        // Check if the datetime value has the "T" character.
        if (str_contains($value[$field_name . '_value'], "T")) {
          // Replace the "T" character with a space.
          $this->dest[$key][$field_name . '_value'] = str_replace('T', ' ', $value[$field_name . '_value']);
        }

        // Check if the datetime value has the ':' character.
        if (!strpos($value[$field_name . '_value'], ":")) {
          // Add ' 00:00:00' to the datetime value.
          $this->dest[$key][$field_name . '_value'] = $value[$field_name . '_value'] . ' 00:00:00';
        }
      }
    }
  }

  /**
   * Preprocessor for user_reference field type.
   *
   * This function is a preprocessor for the user_reference field type. It
   * preprocesses the user reference field by calling the
   * DataMigrationReportCommands::preprocessEntityReference() method with the
   * field and '_uid' as arguments.
   *
   * @param array $field
   *   The user reference field.
   */
  public function userReference(array $field): void {
    $this->preprocessEntityReference($field, '_uid');
  }

  /**
   * Preprocessor for node_reference field type.
   *
   * This function is a preprocessor for the node_reference field type. It calls
   * the preprocessEntityReference() function with the field and '_nid' as
   * arguments.
   *
   * @param array $field
   *   The field being processed.
   */
  public function nodeReference(array $field): void {
    $this->preprocessEntityReference($field, '_nid');
  }

  /**
   * Preprocessor for taxonomy_term_reference field type.
   *
   * This function is a preprocessor for the taxonomy_term_reference field type.
   * It calls the preprocessEntityReference() function with the field and '_tid'
   * as arguments.
   *
   * @param array $field
   *   The field being processed.
   */
  public function taxonomyTermReference(array $field): void {
    $this->preprocessEntityReference($field, '_tid');
  }

  /**
   * Preprocessor for file field type.
   *
   * This function is a preprocessor for the file field type.
   * It calls the preprocessEntityReference() function with the field and '_fid'
   * as arguments.
   *
   * @param array $field
   *   The field being processed.
   */
  public function file(array $field): void {
    $this->preprocessEntityReference($field, '_fid');
  }

  /**
   * Preprocessor for image field type.
   *
   * This function is a preprocessor for the image field type.
   * It calls the preprocessEntityReference() function with the field and '_fid'
   * as arguments.
   *
   * @param array $field
   *   The field being processed.
   */
  public function image(array $field): void {
    $this->preprocessEntityReference($field, '_fid');
  }

  /**
   * Preprocessor for entity reference field types.
   *
   * This function is used to preprocess entity reference fields in the
   * destination data. It changes the field name to match the field name in
   * Drupal 10 and removes the '_target_id' suffix.
   *
   * @param array $field
   *   The field definition array.
   * @param string $column_suffix
   *   The suffix to be appended to the field name.
   */
  public function preprocessEntityReference(array $field, string $column_suffix): void {
    // Get the field name.
    $field_name = $field['name'];

    // Iterate through the destination data array.
    foreach ($this->dest as $key => $value) {
      // Check if the field name with '_target_id' suffix exists in the
      // destination data.
      if (array_key_exists($field_name . '_target_id', $this->dest[$key])) {
        // Change the field name to match the field name in Drupal 7.
        $this->dest[$key][$field_name . $column_suffix] = $this->dest[$key][$field_name . '_target_id'];
        unset($this->dest[$key][$field_name . '_target_id']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping[$field_name . $column_suffix] = $field_name . '_target_id';
      }
    }
  }

  /**
   * Preprocessor for link field type.
   *
   * This function preprocesses link field type by renaming '_uri' to '_url'
   * and '_options' to '_attributes' in the destination array.
   *
   * @param array $field
   *   The field array.
   */
  public function link(array $field): void {
    // Get the field name.
    $field_name = $field['name'];

    // Loop through the destination array.
    foreach ($this->dest as $key => $value) {
      // Rename '_uri' to '_url'.
      if (array_key_exists($field_name . '_uri', $this->dest[$key])) {
        $this->dest[$key][$field_name . '_url'] = $this->dest[$key][$field_name . '_uri'];
        unset($this->dest[$key][$field_name . '_uri']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping[$field_name . '_url'] = $field_name . '_uri';
      }

      // Rename '_options' to '_attributes'.
      if (array_key_exists($field_name . '_options', $this->dest[$key])) {
        $this->dest[$key][$field_name . '_attributes'] = $this->dest[$key][$field_name . '_options'];
        unset($this->dest[$key][$field_name . '_options']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping[$field_name . '_attributes'] = $field_name . '_options';
      }
    }
  }

  /**
   * Preprocessor for user entity type.
   *
   * This function maps the Role ID from the source to the destination if it
   * exists in the mapping table. It also renames the 'roles_target_id' column
   * to 'rid'.
   */
  public function entityUser(): void {
    // Fetch the mapping of source role IDs to destination role IDs.
    $role_mapping = $this->database->select('migrate_map_d7_user_role', 'mur')
      ->fields('mur', ['destid1', 'sourceid1'])
      ->execute()->fetchAllKeyed();

    // Loop through the destination array.
    foreach ($this->dest as $key => $value) {
      // Map the Role ID from the source to the destination if it exists.
      if (array_key_exists('roles_target_id', $this->dest[$key])) {
        if (!empty($this->dest[$key]['roles_target_id'])) {
          if (is_array($this->dest[$key]['roles_target_id'])) {
            foreach ($this->dest[$key]['roles_target_id'] as $role_key => $role_id) {
              if (array_key_exists($role_id, $role_mapping)) {
                $this->dest[$key]['roles_target_id'][$role_key] = $role_mapping[$this->dest[$key]['roles_target_id'][$role_key]];
              }
            }
          }
          else {
            if (array_key_exists($this->dest[$key]['roles_target_id'], $role_mapping)) {
              $this->dest[$key]['roles_target_id'] = $role_mapping[$this->dest[$key]['roles_target_id']];
            }
          }
        }
      }

      // Rename the 'roles_target_id' column to 'rid'.
      if (array_key_exists('roles_target_id', $this->dest[$key])) {
        $this->dest[$key]['rid'] = $this->dest[$key]['roles_target_id'];
        unset($this->dest[$key]['roles_target_id']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping['rid'] = 'roles_target_id';
      }
    }
  }

  /**
   * Preprocessor for node entity type.
   *
   * This preprocessor renames the 'langcode' field to 'language' in the
   * destination array. It also updates the fieldPreprocessMapping array to
   * reflect the change.
   */
  public function entityNode(): void {
    // Loop through the destination array.
    foreach ($this->dest as $key => $value) {
      if (array_key_exists('langcode', $this->dest[$key])) {
        $this->dest[$key]['language'] = $this->dest[$key]['langcode'];
        unset($this->dest[$key]['langcode']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping['language'] = 'langcode';
      }
    }
  }

  /**
   * Preprocessor for taxonomy_term entity type.
   *
   * This function maps the vocabulary ID from the source to the destination
   * if it exists in the mapping table. It also renames the 'description'
   * field to 'description__value' and the 'description__format' field to
   * 'format'.
   */
  public function entityTaxonomyTerm(): void {
    // Fetch the mapping of source vocabulary IDs to destination vocabulary IDs.
    $taxonomy_vocabulary_mapping = $this->database->select('migrate_map_d7_taxonomy_vocabulary', 'mtv')
      ->fields('mtv', ['destid1', 'sourceid1'])
      ->execute()->fetchAllKeyed();

    // Loop through the destination array.
    foreach ($this->source as $key => $value) {
      // If 'description' field is empty, clear the 'format' field if it exists.
      if (array_key_exists('description', $this->source[$key]) && empty($this->source[$key]['description'])) {
        if (array_key_exists('format', $this->source[$key]) && !empty($this->source[$key]['format'])) {
          $this->source[$key]['format'] = '';
        }
      }
    }

    // Loop through the destination array.
    foreach ($this->dest as $key => $value) {
      // Map the vocabulary ID from the source to the destination if it exists.
      if (array_key_exists('vid', $this->dest[$key])) {
        if (!empty($this->dest[$key]['vid']) && array_key_exists($this->dest[$key]['vid'], $taxonomy_vocabulary_mapping)) {
          $this->dest[$key]['vid'] = $taxonomy_vocabulary_mapping[$this->dest[$key]['vid']];
        }
      }

      // Rename the 'description__value' column to 'description'.
      if (array_key_exists('description__value', $this->dest[$key])) {
        $this->dest[$key]['description'] = $this->dest[$key]['description__value'];
        unset($this->dest[$key]['description__value']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping['description'] = 'description__value';
      }

      // Rename the 'description__format' column to 'format'.
      if (array_key_exists('description__format', $this->dest[$key])) {
        $this->dest[$key]['format'] = $this->dest[$key]['description__format'];
        unset($this->dest[$key]['description__format']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping['format'] = 'description__format';
      }
    }
  }

  /**
   * Preprocessor for comment entity type.
   */
  public function entityComment(): void {
    // Loop through the source array.
    foreach ($this->source as $key => $value) {
      if (array_key_exists('pid', $this->source[$key]) && $this->source[$key]['pid'] == 0) {
        $this->source[$key]['pid'] = NULL;
      }
    }

    // Loop through the destination array.
    foreach ($this->dest as $key => $value) {
      if (array_key_exists('entity_id', $this->dest[$key])) {
        $this->dest[$key]['nid'] = $this->dest[$key]['entity_id'];
        unset($this->dest[$key]['entity_id']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping['nid'] = 'entity_id';
      }

      if (array_key_exists('langcode', $this->dest[$key])) {
        $this->dest[$key]['language'] = $this->dest[$key]['langcode'];
        unset($this->dest[$key]['langcode']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping['language'] = 'langcode';
      }
    }
  }

  /**
   * Preprocessor for menu_link_content entity type.
   *
   * This preprocessor renames some columns in the destination array and updates
   * the fieldPreprocessMapping array to reflect the change.
   */
  public function entityMenuLinkContent(): void {
    // Loop through the destination array.
    foreach ($this->dest as $key => $value) {
      // Rename the 'parent' column to 'plid' and pass 0 if no parent.
      if (array_key_exists('parent', $this->dest[$key])) {
        $this->dest[$key]['plid'] = $this->dest[$key]['parent'] ?? 0;
        unset($this->dest[$key]['parent']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping['plid'] = 'parent';
      }

      // Rename the 'link__uri' column to 'link_path'.
      if (array_key_exists('link__uri', $this->dest[$key])) {
        // Convert internal:/ to <front>.
        if ($this->dest[$key]['link__uri'] == 'internal:/') {
          $this->dest[$key]['link__uri'] = '<front>';
        }

        $this->dest[$key]['link_path'] = preg_replace("/\b(?:entity:|base:|internal:\/)\b/", "", $this->dest[$key]['link__uri']);
        unset($this->dest[$key]['link__uri']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping['link_path'] = 'link__uri';
      }

      // Rename the 'title' column to 'link_title'.
      if (array_key_exists('title', $this->dest[$key])) {
        $this->dest[$key]['link_title'] = $this->dest[$key]['title'];
        unset($this->dest[$key]['title']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping['link_title'] = 'title';
      }

      // Rename the 'link__options' column to 'options'.
      if (array_key_exists('link__options', $this->dest[$key])) {
        $this->dest[$key]['options'] = $this->dest[$key]['link__options'];
        unset($this->dest[$key]['link__options']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping['options'] = 'link__options';
      }
    }
  }

  /**
   * Preprocessor for block_content entity type.
   *
   * This preprocessor renames some columns in the destination array and updates
   * the fieldPreprocessMapping array to reflect the change.
   */
  public function entityBlockContent(): void {
    // Loop through the destination array.
    foreach ($this->dest as $key => $value) {
      // Rename the 'body_value' column to 'body'.
      if (array_key_exists('body_value', $this->dest[$key])) {
        $this->dest[$key]['body'] = $this->dest[$key]['body_value'];
        unset($this->dest[$key]['body_value']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping['body'] = 'body_value';
      }

      // Rename the 'body_format' column to 'format'.
      if (array_key_exists('body_format', $this->dest[$key])) {
        $this->dest[$key]['format'] = $this->dest[$key]['body_format'];
        unset($this->dest[$key]['body_format']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping['format'] = 'body_format';
      }
    }
  }

  /**
   * Preprocessor for path_alias entity type.
   *
   * This preprocessor renames some columns in the destination array and updates
   * the fieldPreprocessMapping array to reflect the change.
   */
  public function entityPathAlias(): void {
    // Loop through the destination array.
    foreach ($this->dest as $key => $value) {
      // Remove a leading slash.
      if (array_key_exists('alias', $this->dest[$key])) {
        $this->dest[$key]['alias'] = ltrim($this->dest[$key]['alias'], '/');
      }

      // Rename the 'path' column to 'source' and remove a leading slash.
      if (array_key_exists('path', $this->dest[$key])) {
        $this->dest[$key]['source'] = ltrim($this->dest[$key]['path'], '/');
        unset($this->dest[$key]['path']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping['source'] = 'path';
      }

      // Rename the 'langcode' column to 'language'.
      if (array_key_exists('langcode', $this->dest[$key])) {
        $this->dest[$key]['language'] = $this->dest[$key]['langcode'];
        unset($this->dest[$key]['langcode']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping['language'] = 'langcode';
      }
    }
  }

  /**
   * Preprocessor for entity_reference_revisions field type.
   *
   * This function preprocesses entity_reference_revisions field type by
   * renaming '_target_revision_id' to '_revision_id'
   * and '_target_id' to '_value' in the destination array.
   *
   * @param array $field
   *   The field array.
   */
  public function entityReferenceRevisions(array $field): void {
    // Get the field name.
    $field_name = $field['name'];

    // Loop through the destination array.
    foreach ($this->dest as $key => $value) {
      // Rename '_target_revision_id' to '_revision_id'.
      if (array_key_exists($field_name . '_target_revision_id', $this->dest[$key])) {
        $this->dest[$key][$field_name . '_revision_id'] = $this->dest[$key][$field_name . '_target_revision_id'];
        unset($this->dest[$key][$field_name . '_target_revision_id']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping[$field_name . '_revision_id'] = $field_name . '_target_revision_id';
      }

      // Rename '_target_id' to '_value'.
      if (array_key_exists($field_name . '_target_id', $this->dest[$key])) {
        $this->dest[$key][$field_name . '_value'] = $this->dest[$key][$field_name . '_target_id'];
        unset($this->dest[$key][$field_name . '_target_id']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping[$field_name . '_value'] = $field_name . '_target_id';
      }
    }
  }

}
