<?php

declare(strict_types=1);

namespace Drupal\data_migration_report\Form;

use Drupal\Core\Database\Database;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\State\StateInterface;
use Drupal\migrate_drupal\MigrationConfigurationTrait;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Migrate Upgrade database credential form.
 */
class CredentialForm extends FormBase {

  use MigrationConfigurationTrait;

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * An array of error information.
   *
   * @var array
   */
  protected array $errors = [];

  /**
   * Constructs a new CredentialForm object.
   *
   * @param \Drupal\Core\State\StateInterface $state
   *   The state service.
   */
  public function __construct(StateInterface $state) {
    $this->state = $state;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): CredentialForm|static {
    return new static(
      // Load the service required to construct this class.
      $container->get('state')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'data_migration_report_credentials_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {
    // Get the available database drivers.
    $drivers = $this->getDatabaseTypes();
    // Get the keys of the drivers.
    $drivers_keys = array_keys($drivers);
    // Get the default driver.
    $default_driver = current($drivers_keys);

    $form['database'] = [
      '#type'        => 'details',
      '#title'       => $this->t('Source database'),
      '#description' => $this->t('Provide credentials for the database of the source <strong>Drupal 7</strong> site.'),
      '#open'        => TRUE,
    ];

    $form['database']['driver'] = [
      '#type'          => 'radios',
      '#title'         => $this->t('Database type'),
      '#required'      => TRUE,
      '#default_value' => $this->state->get('data_migration_report_database_driver') ?? $default_driver,
    ];

    // If there is only one available database driver, disable the element.
    if (count($drivers) == 1) {
      $form['database']['driver']['#disabled'] = TRUE;
    }

    // Add driver-specific configuration options.
    foreach ($drivers as $key => $driver) {
      $form['database']['driver']['#options'][$key] = $driver->name();
      $form['database']['settings'][$key]           = $driver->getFormOptions([]);
      unset($form['database']['settings'][$key]['advanced_options']['prefix']['#description']);

      // This form is not rebuilt during submission so #limit_validation_errors
      // is not used. The database and username fields for mysql and pgsql must
      // not be required.
      $form['database']['settings'][$key]['database']['#required'] = FALSE;
      $form['database']['settings'][$key]['username']['#required'] = FALSE;
      $form['database']['settings'][$key]['database']['#states']   = [
        'required' => [
          ':input[name=driver]' => ['value' => $key],
        ],
      ];

      if (!str_ends_with($key, '\\sqlite')) {
        $form['database']['settings'][$key]['username']['#states'] = [
          'required' => [
            ':input[name=driver]' => ['value' => $key],
          ],
        ];
        $form['database']['settings'][$key]['password']['#states'] = [
          'required' => [
            ':input[name=driver]' => ['value' => $key],
          ],
        ];
      }

      $form['database']['settings'][$key]['#prefix']                      = '<h2 class="js-hide">' . $this->t('@driver_name settings', ['@driver_name' => $driver->name()]) . '</h2>';
      $form['database']['settings'][$key]['#type']                        = 'container';
      $form['database']['settings'][$key]['#tree']                        = TRUE;
      $form['database']['settings'][$key]['advanced_options']['#parents'] = [$key];
      $form['database']['settings'][$key]['#states']                      = [
        'visible' => [
          ':input[name=driver]' => ['value' => $key],
        ],
      ];

      // Move the host fields out of advanced settings.
      if (isset($form['database']['settings'][$key]['advanced_options']['host'])) {
        $form['database']['settings'][$key]['host']            = $form['database']['settings'][$key]['advanced_options']['host'];
        $form['database']['settings'][$key]['host']['#title']  = 'Database host';
        $form['database']['settings'][$key]['host']['#weight'] = -1;
        unset($form['database']['settings'][$key]['database']['#default_value']);
        unset($form['database']['settings'][$key]['advanced_options']['host']);
      }

      // Set database default values.
      if ($this->state->get('data_migration_report_database')) {
        foreach ($this->state->get('data_migration_report_database') as $k => $v) {
          if (isset($form['database']['settings'][$key][$k])) {
            $form['database']['settings'][$key][$k]['#default_value'] = $v;
          }
        }
      }
    }

    $form['actions']['#type']  = 'actions';
    $form['actions']['submit'] = [
      '#type'        => 'submit',
      '#value'       => $this->t('Save'),
      '#button_type' => 'primary',
      '#weight'      => 10,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state): void {
    // Retrieve the database driver from the form, use reflection to get the
    // namespace, and then construct a valid database array the same as in
    // settings.php.
    $driver            = $form_state->getValue('driver');
    $drivers           = $this->getDatabaseTypes();
    $reflection        = new \ReflectionClass($drivers[$driver]);
    $install_namespace = $reflection->getNamespaceName();

    $database = $form_state->getValue($driver);
    // Cut the trailing \Install from namespace.
    $database['namespace'] = substr($install_namespace, 0, strrpos($install_namespace, '\\'));
    $database['driver']    = $driver;

    // Validate the driver settings and just end here if we have any issues.
    if ($errors = $drivers[$driver]->validateDatabaseSettings($database)) {
      foreach ($errors as $name => $message) {
        $this->errors[$name] = $message;
      }
    }

    // Get the Drupal version of the source database, so it can be validated.
    $error_key = $database['driver'] . '][database';
    if (!$this->errors) {
      try {
        $this->getConnection($database);
      }
      catch (\Exception $e) {
        $msg = $this->t('Failed to connect to your database server. The server reports the following message: %error.<ul><li>Is the database server running?</li><li>Does the database exist, and have you entered the correct database name?</li><li>Have you entered the correct username and password?</li><li>Have you entered the correct database hostname?</li></ul>', ['%error' => $e->getMessage()]);
        $this->errors[$error_key] = $msg;
      }
    }

    // Display all errors as a list of items.
    if ($this->errors) {
      $form_state->setError($form, $this->t('<h3>Resolve all issues below to continue.</h3>'));
      foreach ($this->errors as $name => $message) {
        $form_state->setErrorByName($name, $message);
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    // Get the value of the 'driver' field from the form state.
    $driver = $form_state->getValue('driver');
    // Store the value of the 'driver' field in the state variable.
    $this->state->set('data_migration_report_database_driver', $driver);
    // Get the value of the 'driver' field from the form state and
    // store it in the state variable.
    $this->state->set('data_migration_report_database', $form_state->getValue($driver));
  }

  /**
   * Returns all supported database driver installer objects.
   *
   * @return \Drupal\Core\Database\Install\Tasks[]
   *   An array of available database driver installer objects.
   */
  public static function getDatabaseTypes(): array {
    // Include the installation API.
    include_once DRUPAL_ROOT . '/core/includes/install.inc';

    // Initialize an empty array to store the database types.
    $database_types = [];

    // Iterate over the installable list of database drivers.
    foreach (Database::getDriverList()->getInstallableList() as $name => $driver) {
      // Add the driver's install tasks to the database types array.
      $database_types[$name] = $driver->getInstallTasks();
    }

    // Return the array of available database driver installer objects.
    return $database_types;
  }

}
